/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 *
 * Credits:
 *   23 Aug 01: Pete Wyckoff <pw@osc.edu>: Architecture-indepdent getdents
 *     and stat changes.
 */


#include <config.h>

/* UNIX INCLUDES */
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_SYSCALL_H
#include <syscall.h>
#endif
#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif
#include <utime.h>
#include <arpa/inet.h>
#include <time.h>
#ifdef HAVE_SYS_STATFS_H
#include <sys/statfs.h>
#endif
#ifdef HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif

#ifdef HAVE_LINUX_TYPES_H
#include <linux/types.h>
#endif
#ifdef HAVE_LINUX_DIRENT_H
#include <linux/dirent.h>
#else
	#ifdef HAVE_DIRENT_H
	#include <dirent.h>
	#endif
#endif
#include <syslog.h>

/* PVFS INCLUDES */
#include <mgr.h>
#include <debug.h>
#include <metaio.h>
#include <dfd_set.h>
#include <pvfs_config.h>

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

static void do_signal(int sig_nr);

/* GLOBALS */
static int (*reqfn[])(int, mreq_p, void *, mack_p) = {
	do_chmod,
	do_chown,
	do_close,
	do_lstat,
	do_noop,
	do_open,
	do_unlink,
	do_shutdown,
	do_noop,
	do_fstat,
	do_rename,
	do_iod_info,
	do_mkdir,
	do_fchown,
	do_fchmod,
	do_rmdir,
	do_access,
	do_truncate,
	do_utime,
	do_getdents,
	do_statfs,
	do_noop,
	do_lookup,
	do_ctime,
	do_link,
	do_readlink,
	do_stat,
};

#ifdef DEBUG
/* reqtest structure only used to print meaningful debug messages */
static char *reqtext[] = {
	"chmod",
	"chown",
	"close",
	"lstat",
	"mount",
	"open",
	"unlink",
	"shutdown",
	"umount",
	"fstat",
	"rename",
	"iod_info",
	"mkdir",
	"fchown",
	"fchmod",
	"rmdir",
	"access",
	"truncate",
	"utime",
	"getdents",
	"statfs",
	"noop",
	"lookup",
	"ctime",
	"link",
	"readlink",
	"stat",
/*** ADD NEW CALLS ABOVE THIS LINE ***/
	"error",
	"error",
	"error",
	"error",
	"error"
};
#endif

fslist_p active_p;
int acc; /* accept socket for mgr */
int debug_on  = 0;
int is_daemon = 1;
int random_base = 0;
int mgr_port = MGR_REQ_PORT;
int default_ssize = DEFAULT_SSIZE;
int iclose_sock; /* used with implicit closes only */
fsinfo_p iclose_fsp; /* used with implicit closes only */
sockset socks;

struct in_addr acc_addr;	/* in network byte order */
struct in_addr acc_mask;	/* in network byte order */

/* PROTOTYPES */

int do_all_implicit_closes(int sock);
int do_implicit_closes(void *v_p);
int check_for_implicit_close(void *v_p);
int invalidate_conn(iod_info *info_p);
int send_req(iod_info iod[], int iods, int base, int pcount, ireq_p req_p, 
    void *data_p, iack_p ack_p);
int send_open_ack(int sock, mreq_p req_p, mack_p ack_p, fsinfo_p fs_p, int cap);
char *dmeta2fname(struct dmeta *d_p);
int get_dmeta(char * fname, dmeta_p dir);
int resv_name(char *);
int startup(int argc, char **argv);
int cleanup(void);
int timeout_fns(void);
static void *do_statusdump(int sig_nr);

void *filter_dirents(void *buf, size_t buflen);


/* FUNCTIONS */
#define MAX_REASONABLE_TRAILER 16384
int main(int argc, char **argv)
{
	int sock, ret, opt, timeout = SOCK_TIME_OUT;
	mreq req;
	mack ack;
	char buf_p[MAX_REASONABLE_TRAILER+2];

	acc_addr.s_addr = htonl(ACC_ADDR);
	acc_mask.s_addr = htonl(ACC_MASK);

#ifdef __RANDOM_BASE__
	random_base = 1;
#endif

	memset(&ack, 0, sizeof(ack));
	while ((opt = getopt(argc, argv,"t:dgrp:a:m:s:")) != EOF) {
		switch (opt) {
			case 't' : timeout = atoi(optarg); break;
			case 'd' : is_daemon=0; break;
			case 'g' : debug_on=1; break;
			case 'r' : random_base=1; break;
			case 'p' : mgr_port= atoi(optarg); break;
			case 'a' :
				if (inet_aton(optarg,&acc_addr) == 0) {
					fprintf(stderr, "accept_addr (-a) %s is invalid\n", optarg);
					exit(-1);
				}
				break;
			case 'm' :
				if (inet_aton(optarg,&acc_mask) == 0) {
					fprintf(stderr,
							  "accept_mask (-m) %s is invalid\n", optarg);
					exit(-1);
				}
				break;
			case 's' :
				ret = atoi(optarg);
				if (ret > 0) default_ssize = ret;
				break;
			default	: break;
		}
	}		
	if ((acc = startup(argc, argv)) == -1) cleanup();

	syslog(LOG_NOTICE, "accept addr %s\n", inet_ntoa(acc_addr));
	syslog(LOG_NOTICE, "accept mask %s\n", inet_ntoa(acc_mask));

	initset(&socks);
	addsockrd(acc, &socks);
	for(;;) /* accept connections and handle them */ {
		if ((ret = check_socks(&socks, timeout)) < 0) {
			/* don't return an error if we got an interrupt */
			if (errno != EINTR) PERROR("check_socks in main");
			continue;
		} else if (ret == 0) {
#if 0
			/* removed by Phil, Oct 15 2002 */
			/* this triggers shutdown requests to iods if enabled, 
			 * would make iods clear out jobs and sockets and reparse
			 * config file
			 */
			timeout_fns();
#endif
			continue;
		}
		while ((sock = nextsock(&socks)) >= 0) {
			/* check if this is the accept socket, and if so do an accept */
			if (sock == acc) {
				struct sockaddr_in sanew;
				socklen_t salen = sizeof(sanew);
				if ((sock = accept(acc,(struct sockaddr *)&sanew,&salen)) < 0) {
					PERROR("accept in main");
					continue;
				}
				if ((sanew.sin_addr.s_addr & acc_mask.s_addr) != acc_addr.s_addr) {
					syslog(LOG_NOTICE,
							 "connect attempt from invalid addr %s\n",
							 inet_ntoa(sanew.sin_addr));
					goto socket_cleanup;
				}
#ifdef ENABLE_TRUSTED_PORTS
				if (ntohs(sanew.sin_port) > 1023) {
					syslog(LOG_NOTICE,
							 "unprivileged connect attempt from %s port %d\n",
							 inet_ntoa(sanew.sin_addr), ntohs(sanew.sin_port));
					goto socket_cleanup;
				}
#endif
				addsockrd(sock, &socks); /* needs to be after accept check */
			}

			/* this will timeout if it does not complete in specified timeframe */
			if (brecv_timeout(sock, &req, sizeof(req), 
				REQUEST_BRECV_TIMEOUT_SECS) <
				(int)sizeof(req)) /* get request */ {
				if (errno == EPIPE) {
					LOG1("socket %d closed\n", sock);
					goto socket_cleanup;
				}
				else {
					PERROR("brecv_timeout in main (new request)");
					goto socket_cleanup;
				}
				continue;
			}
			if (req.majik_nr != MGR_MAJIK_NR) {
				ERR("main: invalid magic number; ignoring request\n");
				goto socket_cleanup;
			}
			if (req.release_nr != PVFS_RELEASE_NR) {
				ERR("main: release number from client (%d.%d.%d) does not match release number of this mgr (%d.%d.%d); returning error\n",
					 req.release_nr / 10000,
					 (req.release_nr / 100) % 10,
					 req.release_nr % 10,
					 PVFS_RELEASE_NR / 10000,
					 (PVFS_RELEASE_NR / 100) % 10,
					 PVFS_RELEASE_NR % 10);
				ERR("main: an administrator needs to verify that all PVFS servers and clients are at the same release number.\n");
				ack.majik_nr   = MGR_MAJIK_NR;
				ack.release_nr = PVFS_RELEASE_NR;
				ack.type       = req.type;
				ack.status     = -1;
				ack.dsize      = 0;
				send_error(-1, EINVAL, sock, &ack);
				goto socket_cleanup;
			}
			if (req.dsize >= MAX_REASONABLE_TRAILER) {
				/* this is ridiculous - throw request out */
				ERR1("main: crazy huge dsize = %d\n; ignoring request\n", 
									 (int) req.dsize);
				goto socket_cleanup;
			}
			if (req.type < 0 || req.type > MAX_MGR_REQ) /* invalid request # */ {
				ERR1("invalid request (%d)", req.type);
				goto socket_cleanup;
			}
			if (req.dsize > 0) /* grab any trailing data */ {
				if (brecv(sock, buf_p, req.dsize) < 0) {
					PERROR("brecv in main - getting trailing data");
					goto socket_cleanup;
				}
				buf_p[req.dsize] = '\0'; /* terminate this if string */
			}

			/* initialize request */
			ack.majik_nr   = MGR_MAJIK_NR;
			ack.release_nr = PVFS_RELEASE_NR;
			ack.type       = req.type; /* everything else is filled in in reqfn() */
			ack.status     = 0;
			ack.dsize      = 0;
			LOG2("req: %s on %d\n", reqtext[req.type], sock);
			ret = (reqfn[req.type])(sock, &req, buf_p, &ack); /* handle request */
			continue;

socket_cleanup:
			close(sock);
			delsock(sock, &socks);
			do_all_implicit_closes(sock);
		}
	} /* end of for(;;) */
	return(0);
} /* end of main() */

int startup(int argc, char **argv)
{
	int fd;

	if (getuid() != 0 && geteuid() != 0) {
		ERR("WARNING: pvfsmgr should be run as root\n");
	}

	if ((fd = new_sock()) < 0) {
		PERROR("pvfsmgr: new_sock");
		return(-1);
	}

	set_sockopt(fd, SO_REUSEADDR, 1);

	if (bind_sock(fd, mgr_port) < 0) {
		PERROR("pvfsmgr: bind_sock");
		return(-1);
	}
	if (listen(fd, MGR_BACKLOG) != 0) {
		PERROR("pvfsmgr: listen");
		return(-1);
	}
	if (is_daemon) {
		int logfd;
#ifndef PVFS_LOG_DIR
		char logname[] = "/tmp/mgrlog.XXXXXX";

		if ((logfd = mkstemp(logname)) == -1) {
#else
		/* new, less obtuse log behavior */
		int logflags = O_APPEND|O_CREAT|O_RDWR;
		char logname[4096];		
		snprintf(logname, 4095, "%s/mgr", PVFS_LOG_DIR);
#ifdef LARGE_FILE_SUPPORT
		logflags |= O_LARGEFILE;
#endif
		if ((logfd = open(logname, logflags, 0700)) == -1) {
#endif
			ERR("couldn't create logfile...continuing...\n");
			close(0); close(1); close(2);
		}
		else {
			fchmod(logfd, 0755);
			dup2(logfd, 2);
			dup2(logfd, 1);
			close(0);
		}
		if (fork()) exit(0); /* fork() and kill parent */
		setsid();
	}	

	active_p = fslist_new(); /* initialize filesystem list */

	/* set up SIGUSR1 handler to spit out debugging info */
	signal(SIGUSR1, (void *)do_statusdump);

	/* set up SIGTERM handler to shut things down */
	signal(SIGTERM, (void *)mgr_shutdown);

	/* set up SIGHUP handler to restart the daemon */
	signal(SIGHUP, (void *)restart);
	
	/* catch SIGPIPE and SIGSEGV signals and log them, on SEGV we die */
	signal(SIGPIPE, do_signal);
	signal(SIGSEGV, do_signal);

	umask(0);
	chdir("/"); /* to avoid unnecessary busy filesystems */

	openlog("mgr", LOG_PID, LOG_ACC_FACILITY);

	return(fd);
} /* end of startup() */

int cleanup()
{
	fslist_cleanup(active_p);
	close(acc);
	exit(0);
	return 0;
} /* end of cleanup() */

int do_noop(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	if (bsend(sock, ack_p, sizeof(mack)) < 0) return -1;
	return 0;
}

int do_chmod(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {

	/* check for reserved name first */
	if (resv_name(data_p) == 0) {
		ack_p->status = md_chmod(req_p, (char *)data_p); /* 0 on success */
		ack_p->eno = errno;
	}
	else {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
	}
	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	return(send_error(ack_p->status, ack_p->eno, sock, ack_p));
}

int do_chown(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {

	/* check for reserved name first */
	if (resv_name(data_p) == 0) {
		ack_p->status          = md_chown(req_p, data_p); /* 0 on success */
		ack_p->eno             = errno;
	}
	else {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
	}
	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	return(send_error(ack_p->status, ack_p->eno, sock, ack_p));
}

/* do_access() - perform the access() operation
 *
 * This call also performs a stat(), so we go ahead and return that
 * data.  The file size included in there is not valid, so we -1 it out.
 *
 * Note that if data_p refers to a symbolic link, then depending upon whether
 * to_follow is set or not we try to
 * follow the link and check whether access is allowed on the target
 * rather than the symbolic link. As such a symbolic link does not have
 * any associated permissions associated with it. ie. it appears with mode
 * 777 on a ls -al.
 */
int do_access(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	/* right now we simply check the file not the path */
	int mode_bits = 0;
	int ret;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}
	/*
	 * do a stat so we can figure out owner and permissions, We follow the link here depending upon 
	 * the value of to_follow in the mreq. This is usually the default for POSIX clients,
	 * library need not do this always (esp. pvfs_readlink, pvfs_symlink etc).
	 */
	if(req_p->req.access.to_follow == FOLLOW_LINK) {
		if ((ret = md_stat(data_p, req_p, &(ack_p->ack.access.meta), FOLLOW_LINK)) != 0) {
			ack_p->ack.access.meta.fs_ino = -1;
			return(send_error(ret, errno, sock, ack_p));
		}
	}
	else {
		if ((ret = md_stat(data_p, req_p, &(ack_p->ack.access.meta), NOFOLLOW_LINK)) != 0) {
			ack_p->ack.access.meta.fs_ino = -1;
			return(send_error(ret, errno, sock, ack_p));
		}
	}
	ack_p->ack.access.meta.u_stat.st_size = -1; /* not valid anyway */
	ack_p->eno = EACCES;
	ack_p->status = -1;
	if (req_p->req.access.mode & R_OK)
		mode_bits |= 004;
	if (req_p->req.access.mode & W_OK)
		mode_bits |= 002;
	if (req_p->req.access.mode & X_OK)
		mode_bits |= 001;
	/* check o, g, and u in that order */
	if (((ack_p->ack.access.meta.u_stat.st_mode & 007) &
		 mode_bits) == mode_bits)
		ack_p->status = 0;
	else if (ack_p->ack.access.meta.u_stat.st_gid == req_p->gid &&
		(((ack_p->ack.access.meta.u_stat.st_mode >> 3) & 007) &
		 mode_bits) == mode_bits)
		ack_p->status = 0;
	else if (ack_p->ack.access.meta.u_stat.st_uid == req_p->uid &&
		(((ack_p->ack.access.meta.u_stat.st_mode >> 6) & 007) &
		 mode_bits) == mode_bits)
		ack_p->status = 0;
	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	return(send_error(ack_p->status, ack_p->eno, sock, ack_p));
} /* end of do_access */

int do_truncate(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int fd, ret;
	fsinfo_p fs_p;
	ireq iodreq;
	fmeta meta;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid,
		req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}

	/* get metadata, check permission to write to file */
	if ((fd = meta_open(data_p, O_RDONLY)) < 0
	|| (meta_read(fd, &meta) < 0)
	|| (meta_close(fd) < 0))
	{
		return(send_error(-1, errno, sock, ack_p));
	}

	/* for the moment permissions aren't checked... */

	/* build request for iods, send it */
	memset(&iodreq, 0, sizeof(iodreq));
	iodreq.majik_nr = IOD_MAJIK_NR;
	iodreq.release_nr = PVFS_RELEASE_NR;
	iodreq.type = IOD_TRUNCATE;
	iodreq.dsize = 0;
	iodreq.req.truncate.f_ino = meta.u_stat.st_ino;
	iodreq.req.truncate.length = req_p->req.truncate.length;
	iodreq.req.truncate.p_stat = meta.p_stat;
	iodreq.req.truncate.part_nr = 0; /* set in send_req() */

	if ((ret = send_req(fs_p->iod, fs_p->nr_iods, meta.p_stat.base, 
			meta.p_stat.pcount, &iodreq, NULL, NULL)) < 0)
	{
		return(send_error(ret, errno, sock, ack_p));
	}	
	if (sock >= 0) /* real socket number, send the ack */ {
		ret = bsend(sock, ack_p, sizeof(mack));
		if (ret < 0) return(-1);
	}
	return(0);
}

/* DO_CTIME() - update ctime of a file.  see also do_utime() function,
 * which updates atime and mtime.
 * Updates the ctime of the target of a symbolic link if it exists!
 */
int do_ctime(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int fd, i, myerr;
	fsinfo_p fs_p;
	fmeta meta;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* we do this simply so we can get the fsinfo for use later */
	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}	

	/* open metadata file */
	if ((fd = meta_open(data_p, O_RDWR)) < 0)
	{
		if (errno != EISDIR) {
			myerr = errno;
			PERROR("do_ctime: meta_open");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}
		else fd = 0; /* needed for meta_access() below...Scott was on
						  * crack when he wrote the fn... */
	}

	/* check write access permissions on file */
	if (meta_access(fd, data_p, req_p->uid, req_p->gid, W_OK) < 0)
	{
		/* if no write permissions, then check ownership */
		if(meta_check_ownership(fd, data_p, req_p->uid, req_p->gid) < 0)
		{
			myerr = errno;
			PERROR("do_ctime: meta_access and meta_check_ownership");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}
	}

	/* see if its a directory or file */
	if ((i = check_pvfs(data_p)) < 1) {
		myerr = errno;
		PERROR("do_ctime: check_pvfs");
		errno = myerr;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* read metadata */
	if (i == 1) {
		if (meta_read(fd, &meta) < 0) {
			myerr = errno;
			PERROR("do_ctime: meta_read");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}

		/* modify metadata */
		meta.u_stat.ctime = req_p->req.ctime.createtime;

		if ((i=meta_write(fd, &meta)) < 0) {
			myerr = errno;
			PERROR("do_ctime: meta_write");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}

		meta_close(fd);
	}
	else { /* it's a directory */
		/* TODO: how can we handle this? */
	}

	/* send ack to client */
	if (sock >= 0) /* real socket number, send the ack */ {
		if (bsend(sock, ack_p, sizeof(mack)) < 0)
			return(-1);
	}

	/* success */
	return(0);
}

/* DO_UTIME() - perform utime() operation.  see also do_ctime, which
 * handles updating creation time in a seperate request.
 * This updates the atime and mtime for the target of the symbolic link
 * file if it exists.
 */
int do_utime(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int fd, i, myerr;
	fsinfo_p fs_p;
	finfo_p f_p;
	fmeta meta;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* we do this simply so we can get the fsinfo for use later */
	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}	

	/* open metadata file */
	if ((fd = meta_open(data_p, O_RDWR)) < 0)
	{
		if (errno != EISDIR) {
			myerr = errno;
			PERROR("do_utime: meta_open");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}
		else fd = 0; /* needed for meta_access() below...Scott was on
						  * crack when he wrote the fn... */
	}

	/* check write access permissions on file */
	if (meta_access(fd, data_p, req_p->uid, req_p->gid, W_OK) < 0)
	{
		/* if no write permissions, then check ownership */
		if(meta_check_ownership(fd, data_p, req_p->uid, req_p->gid) < 0)
		{
			myerr = errno;
			PERROR("do_ctime: meta_access and meta_check_ownership");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}
	}

	/* see if its a directory or file */
	if ((i = check_pvfs(data_p)) < 1) {
		myerr = errno;
		PERROR("do_utime: check_pvfs");
		errno = myerr;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* read metadata */
	if (i == 1) {
		if (meta_read(fd, &meta) < 0) {
			myerr = errno;
			PERROR("do_utime: meta_read");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}

		/* modify metadata */
		meta.u_stat.atime = req_p->req.utime.actime;
		meta.u_stat.mtime = req_p->req.utime.modtime;

		if ((i=meta_write(fd, &meta)) < 0) {
			myerr = errno;
			PERROR("do_utime: meta_write");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}

		meta_close(fd);

		/* if the file is open, record the utime event in the file struct */
		if ((f_p = f_search(fs_p->fl_p, meta.u_stat.st_ino)) != NULL) {
			/* we're interested in WHEN the time was set AND what it was set to */
			f_p->utime_event = time(NULL);
			f_p->utime_modtime = req_p->req.utime.modtime;
		}
	}
	else { /* it's a directory */
		struct utimbuf utb;

		utb.actime = req_p->req.utime.actime;
		utb.modtime = req_p->req.utime.modtime;

		if (utime(data_p, &utb) < 0) {
			myerr = errno;
			PERROR("do_utime: utime");
			errno = myerr;
			return(send_error(-1, errno, sock, ack_p));
		}
	}

	/* send ack to client */
	if (sock >= 0) /* real socket number, send the ack */ {
		if (bsend(sock, ack_p, sizeof(mack)) < 0)
			return(-1);
	}

	/* success */
	return(0);
}

/* do_fstat() - perform fstat() operation -- calls do_stat() to do the
 * work
 */
int do_fstat(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	mreq fakereq;
	fsinfo_p fs_p;
	finfo_p f_p;

	/* find filesystem & open file */
	if (!(fs_p = fs_search(active_p, req_p->req.fstat.meta.fs_ino))
		|| !(f_p = f_search(fs_p->fl_p, req_p->req.fstat.meta.u_stat.st_ino)))
	{
		ack_p->status = -1;
		ack_p->eno  = errno;
		bsend(sock, ack_p, sizeof(mack));
		return(-1);
	}

	/* build a fake stat request, call do_stat() */
	fakereq.majik_nr   = MGR_MAJIK_NR;
	fakereq.release_nr = PVFS_RELEASE_NR;
	fakereq.type       = MGR_STAT;
	fakereq.uid        = req_p->uid;
	fakereq.gid        = req_p->gid;
	fakereq.dsize      = strlen(f_p->f_name);

	return(do_stat(sock, &fakereq, f_p->f_name, ack_p));
} /* end of do_fstat() */

int do_lstat(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret, i;
	fsinfo_p fs_p;
	ireq iodreq;

	/* variables used for calculating correct file size */
	uint64_t real_file_sz, tmp_file_sz, iod_file_sz;
	uint64_t strip_sz, stripe_sz, nr_strips, leftovers;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}	

	/* We dont allow a lstat to follow a symbolic link */
	if ((ret = md_stat(data_p, req_p, &(ack_p->ack.stat.meta), NOFOLLOW_LINK)) != 0) {
		return(send_error(ret, errno, sock, ack_p));
	}

	/* if regular file, get size from iods */
	if (S_ISREG(ack_p->ack.stat.meta.u_stat.st_mode)) {
		iodreq.majik_nr       = IOD_MAJIK_NR;
		iodreq.release_nr     = PVFS_RELEASE_NR;
		iodreq.type 			 = IOD_STAT;
		iodreq.dsize 			 = 0;
		iodreq.req.stat.f_ino = ack_p->ack.stat.meta.u_stat.st_ino;
		if ((ret = send_req(fs_p->iod, fs_p->nr_iods, 
			ack_p->ack.stat.meta.p_stat.base, ack_p->ack.stat.meta.p_stat.pcount, 
			&iodreq, data_p, NULL)) < 0)
		{
			return(send_error(ret, errno, sock, ack_p));
		}	

		strip_sz = ack_p->ack.stat.meta.p_stat.ssize;
		stripe_sz = ack_p->ack.stat.meta.p_stat.ssize * 
						ack_p->ack.stat.meta.p_stat.pcount;
		real_file_sz = 0;

		for (i = 0; i < ack_p->ack.stat.meta.p_stat.pcount; i++) {
			/* i gives us the index into the iods, the ith iod in order */
			iod_file_sz = fs_p->iod[(i + ack_p->ack.stat.meta.p_stat.base) %
			              fs_p->nr_iods].ack.ack.stat.fsize;
			if (iod_file_sz == 0) continue;

			/* the plan is to calculate the file size based on what
			 * this particular iod has.  the largest of these sizes
			 * is the actual file size, taking into account sparse
			 * issues.
			 *
			 * there are four components to the calculation:
			 *
			 * soff = should be 0 since it isn't implemented :)
			 *
			 * nr_strips * stripe_sz = this is a calculation of how many
			 *   complete stripes are present.  note that any partial strip
			 *   or the last complete strip (if there are no partials)
			 *   may not be part of a whole stripe (thus the if() below).
			 *
			 * i * strip_sz = this accounts for full strips that should be
			 *   present on iods that come before us in the iod ordering
			 *
			 * leftovers = the last remaining bytes
			 */
			nr_strips = iod_file_sz / strip_sz;
			leftovers = iod_file_sz % strip_sz;
			
			if (leftovers == 0) {
				nr_strips--;
				leftovers += strip_sz;
			}
			tmp_file_sz = nr_strips * stripe_sz + i * strip_sz +
							  leftovers;
			if (tmp_file_sz > real_file_sz) real_file_sz = tmp_file_sz;
		}

		/* we're adding a 64-bit size field in addition to the st_size 
		 * later i'll remove the addition into st_size probably
		 * i'm leaving it for now
		 */
		ack_p->ack.stat.meta.u_stat.st_size = real_file_sz;
		ack_p->ack.stat.meta.fsize = real_file_sz;
	}
	else if(S_ISLNK(ack_p->ack.stat.meta.u_stat.st_mode)) {
		/* fill up the size of the link file here */
		ack_p->ack.stat.meta.fsize = ack_p->ack.stat.meta.u_stat.st_size;
	}

	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	else send_error(-1, ack_p->eno, sock, ack_p);

	if (!errno) errno = ack_p->eno; /* return the old error */
	return(-1);
} /* end of do_lstat() */

int do_stat(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret, i;
	fsinfo_p fs_p;
	ireq iodreq;

	/* variables used for calculating correct file size */
	int64_t real_file_sz, tmp_file_sz, iod_file_sz;
	int64_t strip_sz, stripe_sz, nr_strips, leftovers;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}	

	/* We allow a stat to follow a symbolic link */
	if ((ret = md_stat(data_p, req_p, &(ack_p->ack.stat.meta), FOLLOW_LINK)) != 0) {
		return(send_error(ret, errno, sock, ack_p));
	}

#ifdef MGR_USE_CACHED_FILE_SIZE
	if ( (f_search(fs_p->fl_p, ack_p->ack.stat.meta.u_stat.st_ino) != NULL) 
		  || ack_p->ack.stat.meta.u_stat.st_size == 0 
		  || ack_p->ack.stat.meta.fsize == 0 )
		{ /*file is closed and size isn't 0.  This is mainly for compatibility with pre-existing metadata */
#endif
			
	  /* if regular file, get size from iods */
			if (S_ISREG(ack_p->ack.stat.meta.u_stat.st_mode)) {
				iodreq.majik_nr       = IOD_MAJIK_NR;
				iodreq.release_nr     = PVFS_RELEASE_NR;
				iodreq.type 			 = IOD_STAT;
				iodreq.dsize 			 = 0;
				iodreq.req.stat.f_ino = ack_p->ack.stat.meta.u_stat.st_ino;
				if ((ret = send_req(fs_p->iod, fs_p->nr_iods, 
										  ack_p->ack.stat.meta.p_stat.base, ack_p->ack.stat.meta.p_stat.pcount, 
										  &iodreq, data_p, NULL)) < 0)
					{
						return(send_error(ret, errno, sock, ack_p));
					}	
				
				strip_sz = ack_p->ack.stat.meta.p_stat.ssize;
				stripe_sz = ack_p->ack.stat.meta.p_stat.ssize * 
					ack_p->ack.stat.meta.p_stat.pcount;
				real_file_sz = 0;
				
				for (i = 0; i < ack_p->ack.stat.meta.p_stat.pcount; i++) {
					/* i gives us the index into the iods, the ith iod in order */
					iod_file_sz = fs_p->iod[(i + ack_p->ack.stat.meta.p_stat.base) %
													fs_p->nr_iods].ack.ack.stat.fsize;
					if (iod_file_sz == 0) continue;
					
					/* the plan is to calculate the file size based on what
					 * this particular iod has.  the largest of these sizes
					 * is the actual file size, taking into account sparse
					 * issues.
					 *
					 * there are four components to the calculation:
					 *
					 * soff = should be 0 since it isn't implemented :)
					 *
					 * nr_strips * stripe_sz = this is a calculation of how many
					 *   complete stripes are present.  note that any partial strip
					 *   or the last complete strip (if there are no partials)
					 *   may not be part of a whole stripe (thus the if() below).
					 *
					 * i * strip_sz = this accounts for full strips that should be
					 *   present on iods that come before us in the iod ordering
					 *
					 * leftovers = the last remaining bytes
					 */
					nr_strips = iod_file_sz / strip_sz;
					leftovers = iod_file_sz % strip_sz;
					
					if (leftovers == 0) {
						nr_strips--;
						leftovers += strip_sz;
					}
					tmp_file_sz = nr_strips * stripe_sz + i * strip_sz +
						leftovers;
					if (tmp_file_sz > real_file_sz) real_file_sz = tmp_file_sz;
				}
				
				/* we're adding a 64-bit size field in addition to the st_size 
				 * later i'll remove the addition into st_size probably
				 * i'm leaving it for now
		 */
				ack_p->ack.stat.meta.u_stat.st_size = real_file_sz;
				ack_p->ack.stat.meta.fsize = real_file_sz;

#ifdef MGR_USE_CACHED_FILE_SIZE
				/*store to metadata if closed*/
				if ( (f_search(fs_p->fl_p, ack_p->ack.stat.meta.u_stat.st_ino) == NULL) ){
					int fd;

					/* get fd for meta file */
					if ((fd = meta_open(data_p, O_RDWR)) < 0) {
						perror("do_stat: meta_open");
						return(-1);
					}

					/* Write metadata back */
					if (meta_write(fd, &(ack_p->ack.stat.meta)) < 0) {
						perror("do_stat: meta_write.");
						return(-1);
					}

					/* Close metadata file */
					if (meta_close(fd) < 0) {
						perror ("do_stat: File close");
						return(-1);
					}
				}
#endif
			}
#ifdef MGR_USE_CACHED_FILE_SIZE
		} /* if open or zero... */
	
	/* otherwise do nothing because it is already saved */

#endif

	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	else send_error(-1, ack_p->eno, sock, ack_p);

	if (!errno) errno = ack_p->eno; /* return the old error */
	return(-1);
} /* end of do_stat() */


int do_lookup(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret;
	fsinfo_p fs_p;
	fmeta meta;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}	

	/* We don't allow a lookup to follow a symbolic link also */
	if ((ret = md_stat(data_p, req_p, &meta, NOFOLLOW_LINK)) != 0) {
		return(send_error(ret, errno, sock, ack_p));
	}
	
	/* even though we copy an entire stat structure, we haven't 
	 * calculated the true size */
	memcpy(&(ack_p->ack.lookup.meta.u_stat), &meta.u_stat, sizeof(meta.u_stat));

	if (!ack_p->status) return(bsend(sock, ack_p, sizeof(mack)));
	else send_error(-1, ack_p->eno, sock, ack_p);

	if (!errno) errno = ack_p->eno; /* return the old error */
	return(-1);
} /* end of do_lookup() */


/* do_statfs()
 */
int do_statfs(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	ireq iodreq;
	fsinfo_p fs_p;
#ifdef HAVE_LINUX_STATFS
	struct statfs sfs;
#else
	/* fall back on statvfs() */
	struct statvfs sfs;
#endif
	int64_t min_free_bytes = 0;
	int64_t total_bytes = 0;
	int i, err;

   /* clear the statfs buffer */
   memset(&sfs, 0, sizeof(sfs));

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, ENOENT, sock, ack_p));
	}

	/* make sure we can talk to the iods */
	if ((fs_p = quick_mount(data_p, req_p->uid,
	req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}

	/* statfs or statvfs on the file locally to get file-related values */
#ifdef HAVE_LINUX_STATFS
	err = statfs(data_p, &sfs);
#else
	err = statvfs(data_p, &sfs);
#endif
	if (err == -1) return (send_error(-1, errno, sock, ack_p));

	ack_p->ack.statfs.tot_files = sfs.f_files;
	ack_p->ack.statfs.free_files = sfs.f_ffree;
#ifdef HAVE_LINUX_STATFS
	ack_p->ack.statfs.namelen = sfs.f_namelen;
#else
	ack_p->ack.statfs.namelen = sfs.f_namemax;
#endif
	
	/* if we didn't get an error by now, then our filename is ok */

	/* send request to all iods for this file system */
	memset(&iodreq, 0, sizeof(iodreq));
	iodreq.majik_nr   = IOD_MAJIK_NR;
	iodreq.release_nr = PVFS_RELEASE_NR;
	iodreq.type       = IOD_STATFS;
	iodreq.dsize      = 0;
	if ((err = send_req(fs_p->iod, fs_p->nr_iods, 0, 
			fs_p->nr_iods, &iodreq, data_p, NULL)) < 0)
	{
		return(send_error(err, errno, sock, ack_p));
	}	

	/* pull data together
    *
	 * Note: we total up the bytes on all iods for the total value, but
	 * for the free value we're going to calculate using:
	 * 
	 * (free space) = (# of iods) * (minimum amount free on any one)
	 *
	 * Although we're lying a bit, it's the best value for giving an
	 * expectation of what could be written to the system.
	 */
	min_free_bytes = fs_p->iod[0].ack.ack.statfs.free_bytes;
	for (i=0; i < fs_p->nr_iods; i++) {
		if (min_free_bytes > fs_p->iod[i].ack.ack.statfs.free_bytes)
			min_free_bytes = fs_p->iod[i].ack.ack.statfs.free_bytes;
		total_bytes += fs_p->iod[i].ack.ack.statfs.tot_bytes;
	}
	ack_p->ack.statfs.tot_bytes = total_bytes;
	ack_p->ack.statfs.free_bytes = min_free_bytes * ((int64_t) fs_p->nr_iods);

	if (sock >= 0) /* real socket number, send the ack */ {
		err = bsend(sock, ack_p, sizeof(mack));
		if (err < 0) return(-1);
	}
	return(0);
}

int do_unlink(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret, fd, is_link = 0;
	fsinfo_p fs_p;
	finfo_p f_p;
	ireq iodreq;
	fmeta meta;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}
	/* md_unlink returns -1 on failure, or an open fd to metadata file on
	 * success.  This is so we can hold on to the inode if we need to.
	 */
	if ((fd = md_unlink(req_p, &meta, (char *)data_p, &is_link)) < 0) {
		return(send_error(fd, errno, sock, ack_p));
	}
	/* only if it is a normal file should this be done */
	if(is_link != 1) {
		if ((f_p = f_search(fs_p->fl_p, meta.u_stat.st_ino)) != 0) {
			/* don't delete this one yet */
			LOG("do_unlink: file open, delaying unlink\n");
			f_p->unlinked = fd;
		}
		else /* need to close FD and do IOD call */ {
			meta_close(fd);
			memset(&iodreq, 0, sizeof(iodreq));
			iodreq.majik_nr   = IOD_MAJIK_NR;
			iodreq.release_nr = PVFS_RELEASE_NR;
			iodreq.type       = IOD_UNLINK;
			iodreq.dsize      = 0;
			iodreq.req.unlink.f_ino = meta.u_stat.st_ino;
			if ((ret = send_req(fs_p->iod, fs_p->nr_iods, meta.p_stat.base, 
				meta.p_stat.pcount, &iodreq, NULL, NULL)) < 0)
			{
				return(send_error(ret, errno, sock, ack_p));
			}	
		}
	}
	if (sock >= 0) /* real socket number, send the ack */ {
		ret = bsend(sock, ack_p, sizeof(mack));
		if (ret < 0) return(-1);
	}
	return(0);
} /* end of do_unlink() */

int do_close(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int i, cnt, myerr = 0;
	fsinfo_p fs_p;
	finfo_p f_p;
	ireq iodreq;
	int64_t max_modtime = 0;

	memset(&iodreq, 0, sizeof(iodreq));

	/* find filesystem & open file */
	if (!(fs_p = fs_search(active_p, req_p->req.close.meta.fs_ino))
		|| !(f_p = f_search(fs_p->fl_p, req_p->req.close.meta.u_stat.st_ino)))
	{
		myerr = errno;
		LOG2("do_close: error finding fs (%Ld) or ino (%Ld)\n",
			req_p->req.close.meta.fs_ino,
			req_p->req.close.meta.u_stat.st_ino);
		return(send_error(-1, errno, sock, ack_p));
	}
	if (--f_p->cnt == 0) /* last open instance; tell IODs to close */ {
		LOG("last reference; closing file\n");
		/* send request to close to IODs */
		memset(&iodreq, 0, sizeof(iodreq));
		iodreq.majik_nr        = IOD_MAJIK_NR;
		iodreq.release_nr      = PVFS_RELEASE_NR;
		iodreq.type            = IOD_CLOSE;
		iodreq.dsize           = 0;
		iodreq.req.close.f_ino = f_p->f_ino;
		/* Tells iod to close all instances of open file */
		iodreq.req.close.cap = -1; 
		
		if (send_req(fs_p->iod, fs_p->nr_iods, f_p->p_stat.base, 
				f_p->p_stat.pcount, &iodreq, NULL, NULL) < 0)
		{
			myerr = errno;
		}

		/* md_close() will pull the atime and mtime from the request and
		 * set them in the metadata file.
		 */
		req_p->req.close.meta.u_stat.atime = time(NULL);
#ifdef USE_OLD_TIME_SYSTEM
		req_p->req.close.meta.u_stat.mtime = time(NULL);
#else

		/* find the latest modification time as returned from the iods */
		for (i = 0, cnt = f_p->p_stat.base;
			  i < f_p->p_stat.pcount; 
			  i++, cnt = (cnt + 1) % fs_p->nr_iods) 
		{
			if (fs_p->iod[cnt].ack.status == 0 && 
				 max_modtime < fs_p->iod[cnt].ack.ack.close.modtime)
			{
					max_modtime = fs_p->iod[cnt].ack.ack.close.modtime;
			}
		}

		/* handle case where utime event occurred after last modification */
		if (f_p->utime_event > max_modtime) {
			req_p->req.close.meta.u_stat.mtime = f_p->utime_modtime;
		}
		else {
			req_p->req.close.meta.u_stat.mtime = max_modtime;
		}
#endif

		if (f_p->unlinked >= 0) /* tell IODs to remove the file too */ {
			memset(&iodreq, 0, sizeof(iodreq));
			iodreq.majik_nr         = IOD_MAJIK_NR;
			iodreq.release_nr       = PVFS_RELEASE_NR;
			iodreq.type             = IOD_UNLINK;
			iodreq.dsize            = 0;
			iodreq.req.unlink.f_ino = f_p->f_ino;
			if (send_req(fs_p->iod, fs_p->nr_iods, f_p->p_stat.base, 
					f_p->p_stat.pcount, &iodreq, NULL, NULL) < 0)
			{
				myerr = errno;
			}
			meta_close(f_p->unlinked); /* close the FD that was kept around */
		}
		else {
			/* call md_close() to update times and such */
#ifdef MGR_USE_CACHED_FILE_SIZE
			md_close(req_p, f_p->f_name, fs_p);
#else
			md_close(req_p, f_p->f_name);
#endif
		}
		f_rem(fs_p->fl_p,f_p->f_ino); /* wax file info */
	}
	else /* just need to update the finfo structure */ {
		dfd_clr(sock, &f_p->socks);
	}
	if (myerr) return(send_error(-1, myerr, sock, ack_p));
	if (sock >= 0 && bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);
	return(0);
}

int do_mount(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	fsinfo_p fs_p;
	int niods, ret=-1;
	static char tmpbuf[256];
	iodtabinfo_p tab_p;
	dmeta dir;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if (get_dmeta(data_p, &dir) < 0) {
		return(send_error(ret, errno, sock, ack_p));
	}
	/* if mounted, OK */
	if (fs_search(active_p, dir.fs_ino)) { 
		if (sock > -1) if (bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);
		return(0);
	}

	/* set up iodtab file name */
	strncpy(tmpbuf, data_p, req_p->dsize);
	strcpy(&tmpbuf[req_p->dsize], TABFNAME);

	/* read iodtab file */
	if ((tab_p = parse_iodtab(tmpbuf)) == NULL) {
		return(send_error(ret, errno, sock, ack_p));
	}

	/* allocate and fill in fsinfo structure */
	if ((fs_p = (fsinfo_p) malloc(sizeof(fsinfo)
		+ sizeof(iod_info) * ((tab_p->nodecount)-1))) == NULL)
	{
		return(send_error(ret, errno, sock, ack_p));
	}
	memset(fs_p, 0, sizeof(fsinfo)+sizeof(iod_info)*((tab_p->nodecount)-1));
	fs_p->nr_iods = tab_p->nodecount;
	fs_p->fs_ino  = dir.fs_ino;
	fs_p->fl_p    = NULL;
	for (niods = 0; niods < tab_p->nodecount; niods++) {
		fs_p->iod[niods].addr = tab_p->iod[niods];
		fs_p->iod[niods].sock = -1;
	}

	if ((ret = fs_add(active_p, fs_p)) < 0) {
		PERROR("fs_add in do_mount");
		return(send_error(ret, errno, sock, ack_p));
	}

	return(0);
}

#define RQ_PSTAT req_p->req.open.meta.p_stat
#define AK_PSTAT ack_p->ack.open.meta.p_stat

int do_open(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret, new_entry=0, new_file=0, myerr;
	fsinfo_p fs_p;
	finfo_p f_p;
	ireq iodreq;

	memset(&iodreq, 0, sizeof(ireq));

	/* NOTE:
	 * new_entry is used to indicate that a file needs to be added into
	 * our list of open files (isn't already in there)
	 *
	 * new_file is used to indicate that we're creating this new file.
	 * -- Rob
	 */

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL) {
		LOG1("do_open: quick_mount: NULL returned for %s\n", (char *) data_p);
		return(send_error(-1, errno, sock, ack_p));
	}

	/* check for pcount > nr_iods and base >= nr_iods */
	if (RQ_PSTAT.pcount > fs_p->nr_iods || RQ_PSTAT.base >= fs_p->nr_iods) {
		return(send_error(-1, EINVAL, sock, ack_p));
	}

	/* if default values are requested, fill them in */
	if (RQ_PSTAT.pcount == -1) RQ_PSTAT.pcount = fs_p->nr_iods; 
	if (RQ_PSTAT.ssize == -1)  RQ_PSTAT.ssize  = default_ssize;

	if (RQ_PSTAT.base == -1) {
		if (random_base) {
			/* pick a random base node number */
			int randbase;
			randbase = rand() % fs_p->nr_iods;
			RQ_PSTAT.base = randbase;
		}
		else RQ_PSTAT.base = 0;
	}

	/* check to see if the file exists already */
	if ((ret = meta_open(data_p, O_RDONLY)) < 0) new_file = 1;
	else meta_close(ret);

	/* md_open() fills in the st_ino structure used below and 
	 * performs sanity checking on the values passed in
	 *
	 * It also creates the metadata file in the specified directory.
	 */
	if ((ret = md_open(data_p, req_p, &(ack_p->ack.open.meta))) < 0) {
		ERR2("do_open: md_open failed: %s, flags = 0x%x\n",
		strerror(errno), req_p->req.open.flag);
		if (new_file && !(req_p->req.open.flag & O_CREAT)) {
		    ERR("do_open: tried to open new file without O_CREAT\n");
		}
		return(send_error(ret, errno, sock, ack_p));
	}

	ack_p->ack.open.meta.fs_ino = fs_p->fs_ino;

	f_p = f_search(fs_p->fl_p,ack_p->ack.open.meta.u_stat.st_ino);

	if (!f_p) /* not open */ {
		new_entry=1;
		if (!(f_p = f_new())) return(send_error(-1, errno, sock, ack_p));
		f_p->f_ino   = ack_p->ack.open.meta.u_stat.st_ino; 
		f_p->p_stat  = AK_PSTAT;
		f_p->f_name  = (char *)malloc(strlen(data_p)+1);
		strcpy(f_p->f_name, data_p);
		dfd_set(sock, &f_p->socks);
	}
	else /* already open */ {
		f_p->cap++;
		f_p->cnt++;
		dfd_set(sock, &f_p->socks);
	}

	f_dump(f_p);

	iodreq.majik_nr        = IOD_MAJIK_NR;
	iodreq.release_nr      = PVFS_RELEASE_NR;
	iodreq.type            = IOD_OPEN;
	iodreq.dsize           = 0;
	iodreq.req.open.f_ino  = f_p->f_ino;
	iodreq.req.open.cap    = f_p->cap;
	iodreq.req.open.flag   = req_p->req.open.flag;
	iodreq.req.open.mode   = S_IRUSR | S_IWUSR; /* permissions on file */
	iodreq.req.open.p_stat = AK_PSTAT;
	iodreq.req.open.pnum   = 0;
	
	if ((ret = send_req(fs_p->iod, fs_p->nr_iods, f_p->p_stat.base, 
			f_p->p_stat.pcount, &iodreq, NULL, NULL)) < 0)
	{
		ERR1("do_open (%s): send_req failed\n", (char *) data_p);
		if (new_file) {
			/* here is a tricky case...we need to remove all the files we
			 * created on the nodes that DID work and we need to remove the
			 * metadata file that md_open() created too...
			 *
			 * We're going to get errors from all this, because something
			 * has happened to one of our IODs.  We're just trying to get
			 * everything back to a sane state as best we can.
			 */
			mreq fakereq;
			mack fakeack;

			myerr = errno;

			ERR1("do_open(%s): cleaning up after failed open\n", (char *) data_p);
			fakereq.majik_nr   = MGR_MAJIK_NR;
			fakereq.release_nr = PVFS_RELEASE_NR;
			fakereq.type       = MGR_CLOSE;
			fakereq.uid        = req_p->uid;
			fakereq.gid        = req_p->gid;
			fakereq.dsize      = 0;
			fakereq.req.close.meta.fs_ino        = fs_p->fs_ino;
			fakereq.req.close.meta.u_stat.st_ino = f_p->f_ino;

			/* hack to ensure that this goes through */
			f_p->unlinked = meta_open(data_p, O_RDONLY); 


			/* here we add the file in so that do_close() can find it */
			f_add(fs_p->fl_p, f_p);
			/* we're closing the file so all the IODs will know it is safe
			 * to unlink it; we're telling them to unlink it at the same
			 * time...
			 */
			do_close(-1, &fakereq, NULL, &fakeack);
			/* do_close() gets rid of our copy of the file info at f_p */

			/* now get rid of the metadata file */
			meta_unlink(data_p);
			errno = myerr;
		}
		return(send_error(ret, errno, sock, ack_p));
	}

	if (send_open_ack(sock, req_p, ack_p, fs_p, f_p->cap) < 0) {
		/* handle cleaning up bad send */
		ERR("do_open: send_open_ack failed\n");
		if (new_entry) f_free(f_p);
		return(-1);
	}
	if (new_entry) f_add(fs_p->fl_p, f_p); /* add file info to list if new */
	return(0);
} /* end of do_open() */

/* SEND_OPEN_ACK() - sends ACK and iod addresses through socket 
 * back to an application that has opened a file
 *
 * note: this call always returns the iod information in an order such
 * that the base iod is listed first on the client side.
 */
int send_open_ack(int sock, mreq_p req_p, mack_p ack_p, fsinfo_p fs_p, int cap)
{
	struct sockaddr saddr;
	int saddrsize, pcnt_remain, nr_infos;
	int64_t expected_size = req_p->req.open.ackdsize;
	struct iovec *vec = NULL;
	int vec_count = 0, i, tot_size = 0;
	char *junk = NULL;
	/*
	 * Look at include/req.h for an explanation of this field,
	 * and how it may improve performance 
	 */
	/* See if the field is reasonable */
	if(expected_size < 0 || expected_size >= MAX_REASONABLE_TRAILER) {
		/* We will have to cleanup the socket forcibly
		 * since the main loop() does not clean up 
		 * even if do_open() returns -1
		 */
		close(sock);
		delsock(sock, &socks);
		do_all_implicit_closes(sock);
		return -1;
	}
	ack_p->eno = 0;
	ack_p->dsize = AK_PSTAT.pcount * sizeof(iod_info);
	ack_p->ack.open.cap = cap;

	saddrsize = sizeof(struct sockaddr);
	if (getsockname(sock, &saddr, &saddrsize) < 0) {
		PERROR("getsockname");
		return(-1);
	}
	ack_p->ack.open.meta.mgr = saddr;

	vec_count = 1; /* mack */
	if (AK_PSTAT.base + AK_PSTAT.pcount <= fs_p->nr_iods) {
		/* only 1 trailer data */
		vec_count++;
	}
	else {
		/* otherwise it's going to take multiple xfers */
		pcnt_remain = AK_PSTAT.pcount;
		if (AK_PSTAT.base) /* send partial list */ {
			vec_count++;
			nr_infos = fs_p->nr_iods - AK_PSTAT.base;
			pcnt_remain -= nr_infos;
		}
		while (pcnt_remain >= fs_p->nr_iods) /* send the whole list */ {
			vec_count++;
			pcnt_remain -= fs_p->nr_iods;
		}
		if (pcnt_remain > 0) /* send list part, starting w/iod 0 */ {
			vec_count++;
		}
	}
	if(expected_size > 0 && expected_size > ack_p->dsize) {
		vec_count++; /* sending junk at the end */
	}
	vec = (struct iovec *)malloc(vec_count * sizeof(*vec));
	if(!vec) {
		PERROR("malloc");
		return (-1);
	}
	i = 0;
	vec[i].iov_base = ack_p; /* mack */
	vec[i++].iov_len = sizeof(mack);
	tot_size += sizeof(mack);
	if (AK_PSTAT.base + AK_PSTAT.pcount <= fs_p->nr_iods) {
		vec[i].iov_base = &fs_p->iod[AK_PSTAT.base];
		vec[i++].iov_len = ack_p->dsize;
		tot_size += ack_p->dsize;
	}
	else {
		/* otherwise it's going to take multiple xfers */
		pcnt_remain = AK_PSTAT.pcount;
		if (AK_PSTAT.base) /* send partial list */ {
			nr_infos = fs_p->nr_iods - AK_PSTAT.base;
			vec[i].iov_base = &fs_p->iod[AK_PSTAT.base];
			vec[i++].iov_len = nr_infos * sizeof(iod_info);
			pcnt_remain -= nr_infos;
			tot_size += nr_infos * sizeof(iod_info);
		}
		while (pcnt_remain >= fs_p->nr_iods) /* send the whole list */ {
			vec[i].iov_base = fs_p->iod;
			vec[i++].iov_len = fs_p->nr_iods * sizeof(iod_info);
			pcnt_remain -= fs_p->nr_iods;
			tot_size += fs_p->nr_iods * sizeof(iod_info);
		}
		if (pcnt_remain > 0) /* send list part, starting w/iod 0 */ {
			vec[i].iov_base = fs_p->iod;
			vec[i++].iov_len = pcnt_remain * sizeof(iod_info);
			tot_size += pcnt_remain * sizeof(iod_info);
		}
	}
	/* Junk at the end */
	if(expected_size > 0 && expected_size > ack_p->dsize) {
		int64_t junk_size = expected_size - ack_p->dsize;
		LOG2("Expected size: %lld, size that I sent: %lld\n", expected_size, ack_p->dsize);
		junk = (char *)malloc(junk_size);
		vec[i].iov_base = junk;
		vec[i++].iov_len = junk_size;
		tot_size += junk_size;
	}
	/* Now send it in one shot */
	if((i=bsendv(sock, vec, vec_count)) < tot_size) {
		ERR1("bsendv managed to send %d bytes\n", i);
		PERROR("Error in bsendv!\n");
		free(vec);
		if(junk) free(junk);
		return -1;
	}
	free(vec);
	if(junk) free(junk);
	return(0);
}

#undef RQ_PSTAT
#undef AK_PSTAT

int do_umount(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int ret;
	fsinfo_p fs_p;
	dmeta dir;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* find filesystem & check for mounted */
	if (get_dmeta(data_p, &dir) < 0) {
		errno = ENOENT;
		return(send_error(-1, ENOENT, sock, ack_p));
	}

	if ((fs_p = fs_search(active_p, dir.fs_ino)) == NULL) {
		errno = ENOENT;
		return(send_error(-1, ENOENT, sock, ack_p));
	}

	/* shut down all IODs */
	if ((ret = stop_iods(fs_p)) < 0) {
		return(send_error(ret, errno, sock, ack_p));
	}
	/* close any open files (?) */

	/* remove entry for filesystem */
	fs_rem(active_p, dir.fs_ino);
	/* send ack if this isn't a "fake" request */
	if (sock > -1 ) if (bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);
	return(0);
}

int do_shutdown(int sock, mreq_p req_p, void *data_p, mack_p ack_p)
{
	forall_fs(active_p, stop_iods);
	LOG("Shut down all iods\n");
	bsend(sock, ack_p, sizeof(mack));
	LOG("Sent ack\n");
	cleanup();
	return(0);
}

int do_rename(int sock, mreq_p req_p, void *data_p, mack_p ack_p)
{
	char *c_p = data_p, temp[MAXPATHLEN], *old_name, *new_name;
	int fd, length;

	/* set pointers to old and new filenames */
	old_name = c_p;
	for (;*c_p;c_p++); /* theres got to be a better way */
	new_name = ++c_p;

	/* check for reserved names first */
	if (resv_name(old_name) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	if (resv_name(new_name) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* check metadat for permissions of old file */
	if ((fd = meta_open(old_name, O_RDONLY)) < 0) {
		PERROR("do_rename: meta_open");
		return(send_error(-1, errno, sock, ack_p));
	}
	
	/* check permissions */
	if (meta_access(fd, old_name, req_p->uid,
		req_p->gid, R_OK | W_OK) < 0) {
		meta_close(fd);
		PERROR("do_rename: meta_access (old name)");
		return(send_error(-1, errno, sock, ack_p));
	}
	if (meta_close(fd) < 0) {
		PERROR("do_rename: meta_close");
		return(send_error(-1, errno, sock, ack_p));
	}
	/* check for write/execute permissions on parent directory */
	strncpy(temp, new_name, MAXPATHLEN);
	length = get_parent(temp);
	/* if length < 0, CWD used, directory permissions not checked */
	if (length >= 0) {
		if (meta_access(0, temp, req_p->uid, req_p->gid, X_OK|W_OK) < 0) { 
			PERROR("do_rename: meta_access (new name)");
			return(send_error(-1, errno, sock, ack_p));
		}
	}

	/* just try to remove the new file and see if it works */
	/* FIXME: Shouldnt we also delete this file on the IOD's? */
   if (unlink(new_name) < 0) {
		if (errno != ENOENT) {
			PERROR("do_rename: unlink");
			return(send_error(-1, errno, sock, ack_p));
		}
	}
	/* move to new name */
   if (rename(old_name, new_name) < 0) {
      PERROR("do_rename: rename");
      return(send_error(-1, errno, sock, ack_p));
   }
	if (sock >= 0) /* real socket number, send the ack */ {
		if (bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);
	}
	return(0); /* no errors, no ack to send */
} /* end of do_rename() */

/* Create a hard or soft link */
int do_link(int sock, mreq_p req_p, void *data_p, mack_p ack_p)
{
	char *c_p = data_p, *target_name, *link_name;
	int soft;

	/* set pointers to target and link filenames */
	link_name = c_p;
	for (;*c_p;c_p++); /* theres got to be a better way */
	target_name = ++c_p;
	soft = req_p->req.link.soft;
	if(soft) {
		LOG2("Try to create soft link from %s->%s\n", link_name, target_name);
	}
	else {
		LOG2("Try to create hard link from %s->%s\n", link_name, target_name);
	}

	/*
	 * Check for reserved names!
	 * FIXME: Should a soft link allow for  non-existent reserved target_name?
	 */
	if (resv_name(link_name) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}
	if (resv_name(target_name) != 0) {
 		ack_p->status = -1;
 		ack_p->eno = ENOENT;
 		return(send_error(-1, errno, sock, ack_p));
	}
	/* Now we try to create the hard/soft link */
	if(soft == 1) {
		/* create a symbolic link */
		if (md_symlink(link_name, target_name, req_p) < 0) {
			PERROR("do_link: md_symlink");
			return send_error(-1, errno, sock, ack_p);
		}
		LOG2("Created soft link from %s->%s\n", link_name, target_name);
	}
	else {
		if (md_link(link_name, target_name, req_p) < 0) {
			PERROR("do_link: md_link");
			return send_error(-1, errno, sock, ack_p);
		}
		LOG2("Created hard link from %s->%s\n", link_name, target_name);
	}
	/* Change the ownership */
	if(lchown(link_name, req_p->req.link.meta.u_stat.st_uid,
		req_p->req.link.meta.u_stat.st_gid) < 0) {
		PERROR("do_link: lchown");
 		return send_error(-1, errno, sock, ack_p);
	}
	if (sock >= 0) /* real socket number, send the ack */ {
		if (bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);
	}
	return(0); /* no errors, no ack to send */
} /* end of do_link() */

/* try to read the link */
int do_readlink(int sock, mreq_p req_p, void *data_p, mack_p ack_p)
{
	char *link_name = NULL, *readlink_name = NULL;
	/* set pointers to target and link filenames */
	link_name = (char *)data_p;

	if(!link_name) {
		ack_p->status = -1;
		ack_p->eno = EINVAL;
		return(send_error(-1, errno, sock, ack_p));
	}
	if (resv_name(link_name) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}
	readlink_name = md_readlink(link_name);
	if(!readlink_name) {
		PERROR("do_readlink: md_readlink");
		return(send_error(-1, errno, sock, ack_p));
	}
	LOG2("readlink of %s yielded %s\n", link_name, readlink_name);
	/* ok great. Now send it back */
	ack_p->dsize = strlen(readlink_name);
	if (sock >= 0) {
		/* Send ack back */
		if (bsend(sock, ack_p, sizeof(mack)) < 0) {
			 free(readlink_name);
			 return(-1);
		}
		/* Also send back the output of the readlink */
		if (bsend(sock, readlink_name, ack_p->dsize) < 0) {
			 free(readlink_name);
			 return(-1);
		}
	}
	free(readlink_name);
	return (0);
}


/* do_iod_info() - returns information on iods
 */
int do_iod_info(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	fsinfo_p fs_p;
#if 0
	dmeta dir;
#endif

#if 0
	if (get_dmeta(data_p, &dir) < 0) {
		return(send_error(-1, ENOENT, sock, ack_p));
	}

	if ((fs_p = fs_search(active_p, dir.fs_ino)) == NULL) {
		return(send_error(-1, ENOENT, sock, ack_p));
	}
#endif
	/* make sure we can talk to the iods */
	if ((fs_p = quick_mount(data_p, req_p->uid, req_p->gid, ack_p)) == NULL)
	{
		return(send_error(-1, errno, sock, ack_p));
	}

	ack_p->ack.iod_info.nr_iods = fs_p->nr_iods;
	ack_p->dsize  = fs_p->nr_iods*sizeof(iod_info);

	if (sock >= 0 ) {
		if (bsend(sock, ack_p, sizeof(mack)) < 0) return(-1);

		/* also send back iod_info 	*/
		if (bsend(sock, fs_p->iod, ack_p->dsize) < 0) return(-1);
	}
	return(0);
}

/* do_mkdir() - makes a new directory in the file system
 */
int do_mkdir(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	char temp[MAXPATHLEN];
	int length;
	dmeta dir;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* check permissions to make this directory */
	strncpy(temp, data_p, MAXPATHLEN);
	length = get_parent(temp);
	if (length >= 0) {
		/* check for permissions to write to directory */
		if (meta_access(0, temp, req_p->uid, req_p->gid, X_OK | W_OK)
			< 0) {
			PERROR("do_mkdir: meta_access");
			return(send_error(-1, errno, sock, ack_p));
		}
	}

	/* get metadata of the dir this is going into */
	if (get_dmeta(data_p, &dir) < 0) {
		errno = ENOENT;
		return(send_error(-1, ENOENT, sock, ack_p));
	}
	
	/* make the new directory */
	if ((ack_p->status = mkdir(data_p, 0775)) < 0) {
		return(send_error(-1, errno, sock, ack_p));
	}
	/* set up the dotfile in the new directory */
	dir.dr_uid = req_p->uid;
	if (!(S_ISGID & dir.dr_mode))
		dir.dr_gid = req_p->gid;
	dir.dr_mode = req_p->req.mkdir.mode | (S_ISGID & dir.dr_mode) | S_IFDIR;

	/* NOTE:
	 * It's important to use md_mkdir() here instead of put_dmeta;
	 * md_mkdir() understands to use the fname portion of the dmeta
	 * structure as part of the name of the directory, so it will get the
	 * sd_path field correct.  put_dmeta() will not.
	 */
	if (md_mkdir(data_p, &dir) < 0) {
		errno = ENOENT;
		return(send_error(-1, ENOENT, sock, ack_p));
	}
	/* send ack to lib */
	if (sock >= 0) /* real socket number, send the ack */ {
		if (bsend(sock, ack_p, sizeof(mack)) < 0)
			return(-1);
	}
	return(0);
} /* end of do_mkdir() */

/* do_fchown() - chowns based on file descriptor
 */
int do_fchown(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	mreq fakereq;
	fsinfo_p fs_p;
	finfo_p f_p;

	/* find filesystem & open file */
	if (!(fs_p = fs_search(active_p, req_p->req.fchown.fs_ino))
	|| !(f_p = f_search(fs_p->fl_p, req_p->req.fchown.file_ino)))
	{
		ack_p->status = -1;
		ack_p->eno  = errno;
		bsend(sock, ack_p, sizeof(mack));
		return(-1);
	}

	/* build a fake chown request, call do_chown() */
	fakereq.majik_nr   = MGR_MAJIK_NR;
	fakereq.release_nr = PVFS_RELEASE_NR;
	fakereq.type       = MGR_CHOWN;
	fakereq.uid        = req_p->uid;
	fakereq.gid        = req_p->gid;
	fakereq.dsize      = strlen(f_p->f_name);
	fakereq.req.chown.owner = req_p->req.fchown.owner;
	fakereq.req.chown.group = req_p->req.fchown.group;

	return(do_chown(sock, &fakereq, f_p->f_name, ack_p));
} /* end of do_fchown() */

/* do_fchmod() - chmods based on file descriptor
 */
int do_fchmod(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	mreq fakereq;
	fsinfo_p fs_p;
	finfo_p f_p;

	/* find filesystem & open file */
	if (!(fs_p = fs_search(active_p, req_p->req.fchmod.fs_ino))
	|| !(f_p = f_search(fs_p->fl_p, req_p->req.fchmod.file_ino)))
	{
		ack_p->status = -1;
		ack_p->eno  = errno;
		bsend(sock, ack_p, sizeof(mack));
		return(-1);
	}
	/* build a fake chmod request, call do_chmod() */
	fakereq.majik_nr   = MGR_MAJIK_NR;
	fakereq.release_nr = PVFS_RELEASE_NR;
	fakereq.type       = MGR_CHMOD;
	fakereq.uid        = req_p->uid;
	fakereq.gid        = req_p->gid;
	fakereq.dsize      = strlen(f_p->f_name);
	fakereq.req.chmod.mode = req_p->req.chmod.mode;

	return(do_chmod(sock, &fakereq, f_p->f_name, ack_p));
} /* end of do_fchmod() */

/* do_rmdir() - removes a directory in the file system
 */
int do_rmdir(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	ack_p->status = md_rmdir(req_p, (char *) data_p);
	ack_p->eno  = errno;
	if (!ack_p->status)
		return(bsend(sock, ack_p, sizeof(mack)));
	bsend(sock, ack_p, sizeof(mack));
	errno = ack_p->eno; /* return the old error */
	return(-1);
} /* end of do_rmdir() */

/* DO_GETDENTS() - opens a directory, seeks to the appropriate location,
 * reads directory data out, and pushes it back through the socket.
 *
 * The common case for this with the kernel module is a request for a
 * single entry.  The common case for this for the library is somewhat
 * larger.  We need to strip out the reserved names from the output
 * here.
 *
 * Client expects a certain number of pvsf_dirent structures, not
 * the actual system dirents.  Offset is an opaque pointer on the
 * client, we use it to find where we were in the directory listing.
 */
int do_getdents(int sock, mreq_p req_p, void *data_p, mack_p ack_p) {
	int i = -1, myerr = 0, ret;
	struct dirent dir;
	struct pvfs_dirent *pdir;
	int num_pdir, opos;
	off_t off;

	/* check for reserved name first */
	if (resv_name(data_p) != 0) {
		ack_p->status = -1;
		ack_p->eno = ENOENT;
		return(send_error(-1, errno, sock, ack_p));
	}

	/* do permission checking */
	if (meta_access(0, data_p, req_p->uid, req_p->gid, R_OK) < 0) {
		ack_p->status = -1;                                             
		ack_p->eno = EPERM;                                             
		return(send_error(-1, errno, sock, ack_p));                     
	}

	/* debugging check */
	if (req_p->req.getdents.length % sizeof(struct pvfs_dirent)) {
		ack_p->status = -1;
		ack_p->eno = EINVAL;
		return(send_error(-1, errno, sock, ack_p));
	}
	/* return buffer to client */
	num_pdir = req_p->req.getdents.length / sizeof(struct pvfs_dirent);
	if (!(pdir = malloc(req_p->req.getdents.length)))
		goto do_getdents_error;

	memset(pdir, 0, req_p->req.getdents.length);
	LOG3("do_getdents: name = %s, offset = %Ld, size = %Ld\n",
		(char *) data_p, req_p->req.getdents.offset, 
		req_p->req.getdents.length);

	if ((i = open(data_p, O_RDONLY)) < 0)
		goto do_getdents_error;

	off = req_p->req.getdents.offset;

	/*
	 * Read into the buffer until we get at least one unfiltered entry,
	 * hit EOF, or get an error of some kind.  Just go one at a time
	 * and copy into pvfs_dirent until it is full.
	 */
	for (opos=0; opos < num_pdir; ) {
		/* seek to our starting position */
		if (lseek(i, off, SEEK_SET) < 0)
			goto do_getdents_error;

		if ((ret = syscall(SYS_getdents, i, &dir, sizeof(dir))) < 0)
			goto do_getdents_error;
		if (ret == 0)
			break;

		if (!resv_name(dir.d_name)) {
			pdir[opos].handle = dir.d_ino;
			pdir[opos].off = off;
			strcpy(pdir[opos].name, dir.d_name);
			++opos;
		}
		off = dir.d_off;
	}

	/* send back the final position */
	ack_p->status = 0;
	ack_p->eno = 0;
	ack_p->ack.getdents.offset = off;
	ack_p->dsize = opos * sizeof(struct pvfs_dirent);

	/* everything went ok, send data */
	bsend(sock, ack_p, sizeof(mack));
	if (opos)
		bsend(sock, pdir, ack_p->dsize);
	free(pdir);
	close(i);
	return(0);

do_getdents_error:
	myerr = errno;
	free(pdir);
	if (i >= 0) close(i);
	ack_p->dsize = 0;
	errno = myerr;
	return(send_error(-1, errno, sock, ack_p));
} /* end of do_getdents() */

/* DO_ALL_IMPLICIT_CLOSES(),
 * DO_IMPLICIT_CLOSES(),
 * CHECK_FOR_IMPLICIT_CLOSE()
 *
 * These functions are used to clean up after applications when they
 * exit.  All files that are open are checked to see if they were opened
 * by the application holding open the socket, and if so the files are
 * "closed" (reference count is decremented and files are closed if
 * count reaches zero).
 */
int do_all_implicit_closes(int sock) {
	iclose_sock = sock;
	forall_fs(active_p, do_implicit_closes);
	return(0);
}

int do_implicit_closes(void *v_p) {
	fsinfo_p fs_p = (fsinfo_p) v_p;
	iclose_fsp = fs_p;
	forall_finfo(fs_p->fl_p, check_for_implicit_close);
	return(0);
}

int check_for_implicit_close(void *v_p) {
	finfo_p f_p = (finfo_p) v_p;
	mreq req;
	ireq iodreq;
	int ret, i;
	int64_t max_modtime;
	
	memset(&iodreq, 0, sizeof(iodreq));

	if (dfd_isset(iclose_sock, &f_p->socks)) {
		LOG2("socket %d found in mask for file %ld\n", iclose_sock, f_p->f_ino);
		if (--f_p->cnt > 0) /* not the last reference */ {
			dfd_clr(iclose_sock, &f_p->socks);
		}
		else /* need to close file on IODs */ {
			LOG1("file %ld implicitly closed\n", f_p->f_ino);
			/* send request to close to IODs */
			iodreq.majik_nr        = IOD_MAJIK_NR;
			iodreq.release_nr      = PVFS_RELEASE_NR;
			iodreq.type            = IOD_CLOSE;
			iodreq.dsize           = 0;
			iodreq.req.close.f_ino = f_p->f_ino;
			/* Tells iod to close all instances of open file */
			iodreq.req.close.cap = -1; 
			
			if ((ret = (send_req(iclose_fsp->iod, iclose_fsp->nr_iods,
				f_p->p_stat.base, f_p->p_stat.pcount, &iodreq, NULL, NULL)) < 0))
			{
				PERROR("send_req failed during implicit close\n");
				return(-1);
			}

			/* find the latest modification time as returned from the iods */
			for (i = 0; i < f_p->p_stat.pcount; i++) {
				if (iclose_fsp->iod[i].ack.status == 0
				|| iclose_fsp->iod[i].ack.eno != ENOENT) {
					if (max_modtime < iclose_fsp->iod[i].ack.ack.close.modtime)
						max_modtime = iclose_fsp->iod[i].ack.ack.close.modtime;
				}
			}
			/* handle case where utime event occurred after last modification */
			if (f_p->utime_event > max_modtime) {
				req.req.close.meta.u_stat.mtime = f_p->utime_modtime;
			}
			else {
				req.req.close.meta.u_stat.mtime = max_modtime;
			}
			req.req.close.meta.u_stat.atime = time(NULL);
			/* store the time values in the metadata file */
#ifdef MGR_USE_CACHED_FILE_SIZE
			md_close(&req, f_p->f_name, iclose_fsp);
#else
			md_close(&req, f_p->f_name);
#endif

			if (f_p->unlinked >= 0) /* tell IODs to remove the file too */ {
				iodreq.majik_nr         = IOD_MAJIK_NR;
				iodreq.release_nr       = PVFS_RELEASE_NR;
				iodreq.type             = IOD_UNLINK;
				iodreq.dsize            = 0;
				iodreq.req.unlink.f_ino = f_p->f_ino;
				if ((ret = (send_req(iclose_fsp->iod, iclose_fsp->nr_iods,
					f_p->p_stat.base, f_p->p_stat.pcount, &iodreq, NULL,
					NULL)) < 0))
				{
					PERROR("send_req failed during implicit close (unlink)\n");
					return(-1);
				}
			}

			f_rem(iclose_fsp->fl_p,f_p->f_ino); /* wax file info */
		}
	}
	return(0);
} /* end of check_for_implicit_close() */

/* stop_iods() - closes connections to idle iods, lets them know that we
 * don't have any files open.  Name is something of a misnomer left over
 * from a time when the manager did actually start and stop the I/O
 * daemons.
 *
 * Returns 0.
 */
int stop_iods(void *v_p) {
	int i;
	ireq iodreq;
	fsinfo_p fs_p = (fsinfo_p) v_p;

	memset(&iodreq, 0, sizeof(iodreq));

	/* shut down all IODs */
	iodreq.majik_nr   = IOD_MAJIK_NR;
	iodreq.release_nr = PVFS_RELEASE_NR;
	iodreq.type       = IOD_SHUTDOWN;
	iodreq.dsize      = 0;
	if (send_req(fs_p->iod, fs_p->nr_iods, 0, fs_p->nr_iods, &iodreq, 
		NULL, NULL) < 0) return(-1);
	for (i = 0; i < fs_p->nr_iods; i++) {
		close(fs_p->iod[i].sock);
		fs_p->iod[i].sock = -1;
	}

	return(0);
} /* end of stop_iods() */

/* send_req() - sends a request to a number of iods
 * iod[]  - pointer to array of iod_info structures
 * iods   - number of iods in array
 * base   - base number for file
 * pcount - number of iods on which the file resides
 * req_p  - pointer to request to send to each iod
 *        this may be modified by send_req() to update values in request
 *        such as the pnum value in an OPEN request
 * data_p - pointer to any request trailing data to be sent; size is
 *        stored in req_p->dsize; may be NULL
 * ack_p  - pointer to iack structure; NOT USED...
 *
 * Returns 0 on success, non-zero on failure.
 */
int send_req(iod_info iod[], int iods, int base, int pcount, ireq_p req_p, 
				void *data_p, iack_p ack_p)
{
	int ret, i;
	int cnt=base, errs=0;

	if (!iod || !req_p) {
		ERR("NULL pointer passed to send_req()\n");
		errno = EINVAL;
		return(-1);
	}

	if (base >= iods || pcount > iods || iods < 1) {
		ERR("Invalid value passed to send_req (base, pcount, or nr_iods)\n");
		errno = EINVAL;
		return(-1);
	}

	for (i = 0; i < pcount;
		cnt = (cnt + 1)%iods, 
		(req_p->type == IOD_OPEN) ? req_p->req.open.pnum++ : 1, /* ugly!  */
		(req_p->type == IOD_TRUNCATE) ? req_p->req.truncate.part_nr++ : 1,
		i++)
	{
		/* clear errno */
		iod[cnt].ack.eno = 0;
		iod[cnt].ack.status = 0;

		if (iod[cnt].sock < 0) /* open the connection first */ {
			/* get socket, connect */
			if ((iod[cnt].sock = new_sock()) == -1) {
				iod[cnt].ack.eno = errno;
				iod[cnt].ack.status = -1;
				PERROR("new_sock"); 
				errs++;
				continue;
			}
#ifdef ENABLE_TRUSTED_PORTS
			/* bind to a privileged port */
			if (bind_sock(iod[cnt].sock, -1) < 0) {
				iod[cnt].ack.eno = errno;
				iod[cnt].ack.status = -1;
					ERR3("error binding port to connect to iod %d (%s:%d)\n", cnt,
					inet_ntoa(iod[cnt].addr.sin_addr), 
					ntohs(iod[cnt].addr.sin_port));
				invalidate_conn(&iod[cnt]);
				errs++;
				continue;
			}
#endif
			/* connect */
			if (connect(iod[cnt].sock, (struct sockaddr *)&(iod[cnt].addr),
				sizeof(iod[cnt].addr)) < 0)
			{
				iod[cnt].ack.eno = errno;
				iod[cnt].ack.status = -1;
				ERR3("error connecting to iod %d (%s:%d)\n", cnt,
					inet_ntoa(iod[cnt].addr.sin_addr), 
					ntohs(iod[cnt].addr.sin_port));
				invalidate_conn(&iod[cnt]);
				errs++;
				continue;
			}	
		}
		if ((ret = bsend(iod[cnt].sock, req_p, sizeof(ireq))) < 0) {
			/* error sending request */
			iod[cnt].ack.eno = errno;
			iod[cnt].ack.status = -1;
			ERR3("error sending request to iod %d (%s:%d)\n", cnt,
				inet_ntoa(iod[cnt].addr.sin_addr), 
				ntohs(iod[cnt].addr.sin_port));
			invalidate_conn(&iod[cnt]);
			errs++;
			continue;
		}
		if (req_p->dsize > 0 && data_p
		&& (ret = bsend(iod[cnt].sock, data_p, req_p->dsize)) < 0) {
			/* error sending trailing data */
			iod[cnt].ack.eno = errno;
			iod[cnt].ack.status = -1;
			ERR3("error sending trailing data to iod %d (%s:%d)\n", cnt,
				inet_ntoa(iod[cnt].addr.sin_addr), 
				ntohs(iod[cnt].addr.sin_port));
			invalidate_conn(&iod[cnt]);
			errs++;
			continue;
		}
		/* timeout if the ack doesn't come back relatively quickly */
		if ((ret = brecv_timeout(iod[cnt].sock, &(iod[cnt].ack), sizeof(iack), 
			REQUEST_BRECV_TIMEOUT_SECS)) < (int)sizeof(iack)) {
			/* error receiving ack */
			iod[cnt].ack.eno = errno;
			iod[cnt].ack.status = -1;
			PERROR("brecv_timeout");
			ERR3("error receiving ack from iod %d (%s:%d)\n", cnt,
				inet_ntoa(iod[cnt].addr.sin_addr), 
				ntohs(iod[cnt].addr.sin_port));
			invalidate_conn(&iod[cnt]);
			errs++;
			continue;
		}
	} /* end of forall iods */
	for (i=0, cnt=base; i < iods; i++, cnt=(cnt+1)%iods) {
		if (iod[cnt].ack.status) {
			ERR2("error received from iod %d: %s\n", cnt,
				strerror(iod[cnt].ack.eno));
			errno = iod[cnt].ack.eno;
			return(iod[cnt].ack.status);
		}
	}
	if (errs) {
		errno = EIO;
		return(-1);
	}
	return(0);
} /* end of send_req() */

/* invalidate_conn() - closes an iod socket and removes references to it
 *
 * Returns -1 on obvious error, 0 on success.
 */
int invalidate_conn(iod_info *info_p)
{
	if (!info_p) {
		ERR("NULL iod_info pointer passed to invalidate_conn()\n");
		return(-1);
	}
	close(info_p->sock);
	info_p->sock = -1;
	return(0);
}

/* send_error() - sends an error ack back to an app if a valid
 * socket number (>= 0) is given, and returns -1.
 *
 * sets errno to the value passed in as p_errno
 */
int send_error(int status, int p_errno, int sock, mack_p ack_p) 
{
	if (sock >= 0) {
		LOG("*** SENDING ERROR ACK ***\n");
		ack_p->status = status;
		ack_p->eno  = p_errno;
		ack_p->dsize  = 0;
		bsend(sock, ack_p, sizeof(mack));
	}
	else {
		LOG("*** ERROR - NOT SENDING ACK ***\n");
	}
	errno = p_errno; /* preserve original errno */
	return(-1);
} /* end of send_error() */

/* quick_mount() - checks if file system holding a given file is mounted;
 * if not, it mounts it
 */
fsinfo_p quick_mount(char *fname, int uid, int gid, mack_p ack_p) {
	int ret;
	static mreq fakereq;
	fsinfo_p fs_p;
	dmeta dir;

	memset(&dir, 0, sizeof(dir));
	memset(&fakereq, 0, sizeof(fakereq));

	/* fill in fake request for do_mount() */
	if (get_dmeta(fname, &dir) < 0) {
		ERR1("quick_mount: get_dmeta (%s): ", fname);
		PERROR("");
		return(NULL);
	}

	if ((fs_p = fs_search(active_p, dir.fs_ino)) != 0) return(fs_p);

	LOG1("quick_mount: mounting %s\n", dir.rd_path);
	/* need a / on the end of the root path - hack it - WBL */
	/* this trashes the sd_path, but that shouldn't matter */
	*(dir.sd_path-1) = '/';
	*(dir.sd_path) = '\0';
	fakereq.type  = MGR_MOUNT;
	fakereq.uid   = uid;
	fakereq.gid   = gid;
	fakereq.dsize = strlen(dir.rd_path);
	/* mount fs */
	if ((ret = do_mount(-1, &fakereq, dir.rd_path, ack_p)) < 0) {
		PERROR("quick_mount: do_mount");
		return(NULL);
	}
	/* find filesystem & open file */
	return (fs_search(active_p, dir.fs_ino));
} /* end of quick_mount() */

int timeout_fns()
{
	forall_fs(active_p, stop_idle_iods);
	return(0);
}

int stop_idle_iods(void *v_p)
{
	int ret, ret2;
	fsinfo_p fs_p = (fsinfo_p) v_p;

	if (flist_empty(fs_p->fl_p)) {
		LOG("Closing idle iod connections\n");
		ret = stop_iods(v_p);
		ret2 = fs_rem(active_p, fs_p->fs_ino);
		return(MIN(ret, ret2));
	}
	return(0);
} /* end of stop_idle_iods() */

void *restart(int sig_nr)
{
	LOG("pvfsmgr: received HUP signal\n");

	forall_fs(active_p, stop_iods);
	LOG("Shut down all iods\n");
	fslist_cleanup(active_p);
	active_p = fslist_new(); /* initialize new filesystem list */
	LOG("Cleared filesystem list\n");
	signal(SIGHUP, (void *)restart);

	return(0);
} /* end of restart() */

void *mgr_shutdown(int sig_nr)
{
	forall_fs(active_p, stop_iods);
	LOG("Closed all iod connections\n");
	cleanup();
	return(0);
}

/* do_signal() - catch signals, reset handler, die on SIGSEGV */
static void do_signal(int sig_nr)
{
	ERR1("do_signal: got signal %d\n", sig_nr);
	if (sig_nr == SIGPIPE) {
		signal(sig_nr, (void *)do_signal);
		return;
	}

	ERR("\nOPEN FILES:\n");
	fslist_dump(active_p);
	cleanup();
}

/* do_statusdump() - spit out good info, reset signal handler */
static void *do_statusdump(int sig_nr)
{
	ERR("\nopen files:\n");
	fslist_dump(active_p);

	printf("read set:\n");
	dfd_dump(&socks.read);
	printf("write set:\n");
	dfd_dump(&socks.write);
	printf("tmp_read set:\n");
	dfd_dump(&socks.tmp_read);
	printf("tmp_write set:\n");
	dfd_dump(&socks.tmp_write);

	fflush(stdout);
	fflush(stderr);

	signal(sig_nr, (void *)do_statusdump);
	
	return (void *)0;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 *
 * vim: ts=3
 */
